package MT::Plugin::TwoFactorAuthentication;

use strict;
use MT::Plugin;
use utf8;
use Carp qw(longmess);
use base qw( MT::Plugin );

our $PLUGIN_NAME = 'TwoFactorAuthentication(Trial)';
our $VERSION = '1.13';

use TwoFactorAuthentication::Util qw( plugin );

use MT;
my $plugin = MT::Plugin::TwoFactorAuthentication->new({
    id => 'twofactorauthentication',
    key => __PACKAGE__,
    name => "<__trans phrase='$PLUGIN_NAME'>",
    version => $VERSION,
    description => "<__trans phrase='This plugin enables Two-factor authentication.'>",
    author_name => 'Curlybracket, Inc.',
    author_link => 'https://www.curlybracket.co.jp',
    l10n_class => 'TwoFactorAuthentication::L10N',
    system_config_template => \&system_config_template,
    settings => new MT::PluginSettings([
        ['enabled', { Default => 0, Scope => 'system' }],
        ['site_name', { Default => $ENV{HTTP_HOST} || 'Movable Type', Scope => 'system' }],
        ['two_step_enabled', { Default => 0, Scope => 'system' }],
    ]),
    registry => {
        config_settings => {
        },
        object_types => {
            author => {
                tfa_enabled => 'boolean meta',
                tfa_secret_key => 'string meta',
                tfa_recovery_codes => 'array meta',
            },
        },
        applications => {
            cms => {
                methods => {
                    tfa_verify => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::verify',
                    tfa_do_verify => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::do_verify',
                    tfa_regenerate_secret_key => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::regenerate_secret_key',
                    tfa_view_recovery_codes => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::view_recovery_codes',
                    tfa_generate_recovery_codes => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::generate_recovery_codes',
                    tfa_enable => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::enable',
                    tfa_do_enable => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::do_enable',
                    tfa_disable => '$TwoFactorAuthentication::TwoFactorAuthentication::CMS::disable',
                },
            },
        },
        callbacks =>  {
            'init_app' => {
                priority => 10,
                handler => \&init_app,
            },
            'template_source.edit_author' => \&hdlr_edit_author_source,
            'template_param.edit_author' => \&hdlr_edit_author_param,
            'template_param.login' => \&hdlr_login_param,
        },
    },
});

MT->add_plugin($plugin);

sub instance { $plugin; }

sub is_mt70_or_later {
    my ($major, $minor) = split(/\./, MT->version_number);
    return ($major == 7 && $minor >= 0) || ($major > 7);
}

sub doLog {
    my ($msg) = @_; 
    return unless defined($msg);

    require MT::Log;
    my $log = MT::Log->new;
    $log->message($msg) ;
    $log->save or die $log->errstr;
}

sub system_config_template {
    my $tmpl = <<'EOT';
    <mtapp:setting
        id="enabled"
        label="<__trans phrase="Enabled">"
        hint="<__trans phrase="Enable plugin.">"
        show_hint="1">
        <input type="checkbox" name="enabled" id="enabled" value="1"<mt:if name="enabled"> checked="checked"</mt:if> />
    </mtapp:setting>
    
    <mtapp:setting
        id="site_name"
        label="<__trans phrase="Site name">"
        hint="<__trans phrase="This site name will be displayed on the authentication app.">"
        show_hint="1">
        <input type="text" class="text full" name="site_name" id="site_name" value="<mt:var name="site_name">" />
    </mtapp:setting>
EOT
}

sub enabled {
    my $plugin = shift;
    my ($blog_id) = @_;
    my %plugin_param;

    $plugin->load_config(\%plugin_param, 'system');
    my $value = $plugin_param{enabled};
    return $value;
}

sub site_name {
    my $plugin = shift;
    my ($blog_id) = @_;
    my %plugin_param;

    $plugin->load_config(\%plugin_param, 'system');
    my $value = $plugin_param{site_name};
    return $value;
}

sub two_step_enabled {
    my $plugin = shift;
    my ($blog_id) = @_;
    my %plugin_param;

    $plugin->load_config(\%plugin_param, 'system');
    my $value = $plugin_param{two_step_enabled};
    return $value;
}

sub hdlr_edit_author_source_v7 {
    my ($cb, $app, $tmpl_ref) = @_;

    return 1 unless $plugin->enabled;

    my $target = quotemeta(<<'TMPL');
  <fieldset id="author-preferences-settings" class="form-group">
TMPL
    my $tmpl = $plugin->load_tmpl('edit_author_v7.tmpl');
    my $replacement = $tmpl->text;
    $$tmpl_ref =~ s!($target)!$replacement$1!;

    return 1;
}

sub hdlr_edit_author_source_v6 {
    my ($cb, $app, $tmpl_ref) = @_;

    return 1 unless $plugin->enabled;

    my $target = quotemeta(<<'TMPL');
  <h2><__trans phrase="Preferences"></h2>
TMPL
    my $tmpl = $plugin->load_tmpl('edit_author.tmpl');
    my $replacement = $tmpl->text;
    $$tmpl_ref =~ s!($target)!$replacement$1!;

    return 1;
}

sub hdlr_edit_author_source {
    my ($cb, $app, $tmpl_ref) = @_;
    return is_mt70_or_later() 
        ? hdlr_edit_author_source_v7(@_)
        : hdlr_edit_author_source_v6(@_);
}

sub hdlr_edit_author_param {
    my ($cb, $app, $param, $tmpl) = @_;

    return 1 unless $plugin->enabled;
    my $user = $app->user
        or return 1;
    my $id = $app->param('id')
        or return 1;
    if($id != $user->id) {
        $param->{tfa_other_user} = 1;
    }

    my $type = $app->param('_type');
    my $class = $app->model($type);
    my $obj = $class->load($id)
        or return 1;

    $param->{tfa_enabled} = $obj->tfa_enabled();
    $param->{tfa_secret_key} = $obj->tfa_secret_key();
    return 1;
}

sub hdlr_login_param {
    my ($cb, $app, $param, $tmpl) = @_;

    return 1 unless $plugin->enabled;

    my $query_params = $param->{query_params};
    if(defined $query_params && ref $query_params eq 'ARRAY') {
        @$query_params = grep { $_->{name} ne 'totp' } @$query_params;
        $param->{query_params} = $query_params;
    }

    return 1;
}

sub init_app {
    my ( $cb, $app ) = @_;

    my $plugin = $app->component('TwoFactorAuthentication')
      or return 1;
    return 1 unless $plugin->enabled;
    return 1 if $app->{tfa_init};

    require MT::App;
    no strict 'refs';
    no warnings 'redefine';
    my $saved_pre_run = \&MT::App::pre_run;
    my $_pre_run      = sub {
        my $app = shift;
        my $e;
        return 1 unless $app->isa('MT::App::CMS');
        eval {
            $saved_pre_run->( $app, @_ );
        };
        if($@) {
            $e = $@;
            if($@ !~ m!^PASSWORD_LIFETIME_EXPIRED!) {
                die;
            }
        }
        pre_run($app);
        if($e && !_is_excluded_mode($app)) {
            die $e;
        }
        $app->{redirect} = undef;
        return 1;
    };
    *MT::App::pre_run = $_pre_run;

    my $saved_login = \&MT::App::login;
    my $_login = sub {
        my $app = shift;
        my ($author, $new_login) = $saved_login->($app, @_);
        if($author && $app->isa('MT::App::CMS')) {
            my $sess = $app->{session}
                or return ($author, $new_login);
            if($new_login) {
                $sess->set('tfa_verified', 0);
                $sess->save;
            }
        }
        return ($author, $new_login);
    };
    *MT::App::login = $_login;

    $app->{tfa_init} = 1;

    return 1;
}

sub is_tfa_enabled {
    my ($plugin, $app, $now ) = @_;

    return unless $plugin->enabled;

    return unless $app->can('user');
    my $user = $app->user
      or return;

    return unless $user->tfa_enabled;

    return 1;
}

sub _is_excluded_mode {
    my ($app) = @_;
    my $mode = $app->mode;
    my @excluded_modes = (
        'tfa_verify',
        'tfa_do_verify',
        'change_password',      # for PasswordLifetime
        'do_change_password',   # for Passwordlifetime
    );
    my @match = grep { $_ eq $mode } @excluded_modes;
    return scalar @match ? 1 : 0;
}

sub pre_run {
    my ($app) = @_;

    return 1 unless $app->isa('MT::App::CMS');

    my $plugin = $app->component('TwoFactorAuthentication')
      or return 1;

    my $verified = 0;
    my $sess = $app->{session};
    if($sess) {
        $verified = $sess->get('tfa_verified');
    }

    if (
        $app->{requires_login}
            && !$verified
            && !_is_excluded_mode($app)
      )
    {
        return 1 if $app->config('DisableTwoFactorAuthentication');
        return 1 unless $plugin->is_tfa_enabled($app);

        my $x_request_with = $app->get_header('X-Requested-With') || '';
        if($sess && $x_request_with ne 'XMLHttpRequest') {
            my $mode = $app->mode;
            if($mode && ($mode !~ m/delete/)) {
                my $q = $app->{query};
                my @query;
                my @query_keys = grep {
                    ( $_ ne 'username' )
                        && ( $_ ne 'password' )
                        && ( $_ ne 'submit' )
                        && ( $mode eq 'logout' ? ( $_ ne '__mode' ) : 1 )
                } $q->param;
                for my $query_key (@query_keys) {
                    my @vals = $q->param($query_key);
                    for my $val (@vals) {
                        push @query, { name => $query_key, value => $val };
                    }
                }
                $sess->set('tfa_query_params', \@query);
                $sess->save;
            }
        }

        my $blog = $app->can('blog') ? $app->blog : undef;
        $app->redirect(
            $app->uri(
                mode => 'tfa_verify',
                args => { blog_id => $blog ? $blog->id : 0, },
            )
        );
        die 'DO_TWO_FACTOR_AUTHENTICATION';
    }

    return 1;
}

sub reset_config {
    my $plugin = shift;
    my ($scope) = @_;
    if($scope && $scope eq 'system') {
        require MT::Author;
        my $driver = MT::Author->driver;
        if($driver) {
            $driver->sql([
                'delete from mt_author_meta where author_meta_type like "tfa_%"',
            ]);
        }
    }
    MT->instance->reboot if MT->instance->can('reboot');
    return $plugin->SUPER::reset_config(@_);
}

sub save_config {
    my $plugin = shift;
    MT->instance->reboot if MT->instance->can('reboot');
    return $plugin->SUPER::save_config(@_);
}

1;
__END__

