package TwoFactorAuthentication::Auth;
use strict;
use warnings;
use utf8;
use Carp;
use base qw( Class::Accessor::Fast MT::ErrorHandler );

use TwoFactorAuthentication::Util qw( plugin );
use Digest::SHA qw( hmac_sha1_hex );
use MT::Util qw( encode_url );

__PACKAGE__->mk_accessors(qw( user app ));

use constant TIME_STEP => 30;

sub _trans {
    return plugin->translate(@_);
}

sub new {
    my $class = shift;
    my $self = bless { }, $class;
    return $self->init(@_);
}

sub init {
    my $self = shift;
    my %param = @_;

    if(!$param{user}) {
        croak "User is required.";
    }
    $self->user($param{user});

    if(!$param{app}) {
        croak "App is required.";
    }
    $self->app($param{app});

    if(exists $param{secret}) {
        my $secret = $param{secret} || $self->generate_secret;
        $self->user->tfa_secret_key($secret);
        $self->user->save
            or croak "Failed to save user.";
    }

    return $self;
}

sub generate_secret {
    my $self = shift;
    my @chars = ("A" .. "Z", "2" .. "7");
    my $length = scalar(@chars);
    my $secret = "";
    for (my $i = 0; $i < 16; $i++) {
        $secret .= $chars[rand($length)];
    }
    return $secret;
}

sub _decode_base32 {
    my ($val) = @_;
    $val =~ tr|A-Z2-7|\0-\37|;
    $val = unpack('B*', $val);
    $val =~ s/000(.....)/$1/g;
    my $len = length($val);
    $val = substr($val, 0, $len & ~7) if $len & 7;
    $val = pack('B*', $val);
    return $val;
}

sub calculated_totp {
    my $self = shift;
    my $secret = $self->user->tfa_secret_key
        or return $self->errtrans('No secret key.');
    my $padded_time = sprintf("%016x", int(time() / TIME_STEP));
    my $data = pack('H*', $padded_time);
    my $key = _decode_base32($secret);
    my $hmac = hmac_sha1_hex($data, $key);
    my $offset = hex(substr($hmac, - 1));
    my $encrypted = hex(substr($hmac, $offset * 2, 8)) & 0x7fffffff;
    my $token = $encrypted % 1000000;
    return sprintf("%06d", $token);
}

sub errtrans {
    my $self = shift;
    return $self->error(_trans(@_));
}

sub is_valid_totp {
    my $self = shift;
    my ($totp) = @_;

    return $self->errtrans('Enter veirfication code.') unless $totp;
    $totp =~ s!\s+!!g;

    my $calculated_totp = $self->calculated_totp();
    if ($totp eq $calculated_totp) {
        $self->set_session('tfa_verified', 1);
        return 1;
    }

    # Check recovery code
    my $recovery_codes = $self->user->tfa_recovery_codes || [ ];
    foreach my $recovery_code (@$recovery_codes) {
        if (!$recovery_code->{used} && $totp eq $recovery_code->{code}) {
            $recovery_code->{used} = 1;
            $self->user->tfa_recovery_codes($recovery_codes);
            $self->user->save;
            $self->set_session('tfa_verified', 1);
            return 1;
        }
    }

    return $self->errtrans('Verification code does not match.');
}

sub set_session {
    my $self = shift;
    my ($key, $value) = @_;

    return unless $key;

    my $sess = $self->app->{session}
        or return;
    $sess->set($key, $value);
    $sess->save;

    return;
}

sub secret_key {
    my $self = shift;
    my $secret = $self->user->tfa_secret_key
        or return $self->errtrans('No secret key.');
    return $secret;
}

sub qrcode_src {
    my $self = shift;

    my $site_name = plugin->site_name || 'Movable Type';
    my $secret = $self->secret_key()
        or return;
    my $url = 'otpauth://totp/' . $self->user->name . '?secret='. $secret . '';
    if ($site_name) {
        $url .= '&issuer=' . $site_name;
    }
    my $encoded_url = encode_url($url);
    my $qrcode_src = 'https://chart.googleapis.com/chart?chs=200x200&chld=M|0&cht=qr&chl='. $encoded_url . '';

    return $qrcode_src;
}

sub qrcode_url {
    my $self = shift;
    my $site_name = plugin->site_name || 'Movable Type';
    my $secret = $self->secret_key()
        or return;
    my $url = 'otpauth://totp/' . encode_url($self->user->name) . '?secret='. $secret . '';
    if ($site_name) {
        $url .= '&issuer=' . encode_url($site_name);
    }
    return $url;
}

sub enabled {
    my $self = shift;
    my $enabled = $self->user->tfa_enabled;
    return $enabled;
}

sub make_recovery_code {
    my $self = shift;
    my @alpha = ( 'a' .. 'z', 'A' .. 'Z', 0 .. 9 );
    my $token = join '', map $alpha[ rand @alpha ], 1 .. 10;
    return $token;
}

sub generate_recovery_codes {
    my $self = shift;
    # Generate recovery codes
    my @recovery_codes = map { { code => $self->make_recovery_code(), used => 0 } } (1..10);
    return \@recovery_codes;
}

sub recovery_codes {
    my $self = shift;
    return $self->user->tfa_recovery_codes;
}

sub enable {
    my $self = shift;

    $self->user->tfa_enabled(1);
    $self->user->tfa_recovery_codes($self->generate_recovery_codes);
    $self->user->save
        or return $self->errtrans('Failed to enable.');
    return 1;
}

sub disable {
    my $self = shift;

    $self->user->tfa_enabled(0);
    $self->user->tfa_secret_key(undef);
    $self->user->tfa_recovery_codes(undef);
    $self->user->save
        or return $self->Error('Failed to disable.');
}

sub regenerate_recovery_codes {
    my $self= shift;
    $self->user->tfa_recovery_codes($self->generate_recovery_codes);
    $self->user->save
        or return $self->errtrans('Failed to regenerate recovery codes.');
    return 1;
}

1;
__END__
